/**
 *	comsub  --  communication subroutines
 *		Copyright (C) Kazuro Furukawa, 1988-2004
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa  Nov.9.1988.  (rwrite)
 *      K.F  Apr.13.1989.  converted from rwrite (recho)
 *	K.F  Mar.19.1991.  recho v3.4.2
 *	K.F  Mar.24.1992.(2)  comsub
 *	K.F  Sep.14.1992.  add option 3 (bit 2)
 *	Modifications
 *	K.F  Nov.26.1992.  close failed socket correctly
 *	K.F  Mar.2.1993.   move error code definition to comsub.h
 *	K.F  Jun.10.1993.  add s2_bopen() for broadcasting
 *	K.F  Feb.21.1994.  add s2_preopen() and s2_postopen
 *	K.F  Mar.3.1994.   add s2_scopen()
 *	K.F  Feb.21.1996.  ansi C
 *	K.F  Jun.29.1996.  fix bug in s2_currtime
 *	K.F  Dec.11.1996.  add s2_currtime1()
 *	K.F  Aug.26.1997.  fix bug in s2_preopen at host address
 *	K.F  May.11.1999.  with NONETDB mainly for vxworks
 *	K.F  May.19.1999.  sleep functions
 *	K.F  Jul.21.1999.  check hostname argument in s2_preopen
 *	K.F  Apr.2.2000.   support for windows/winsock
 *	K.F  Aug.17.2003.  change usage of gettimeofday for new standards
 *
 *	This module is a re-writing of zechosub.
 *	This module defines basic rouines for socket communication.
 *	 int s2_open( portnum, service, protocol, hostname, option )
 *	 int s2_scopen( destination, protocol, option )
 *	 int s2_close( sd, option )
 *	 int s2_send( sd, buffer, buflen, option )
 *	 int s2_receive( sd, buffer, buflen, option )
 *	This module also defines rouines for broadcast communication.
 *	 int s2_bopen( portnum, service, protocol, hostname, option )
 *	 int s2_bsend( sd, buffer, buflen, option )
 *	And miscelaneous routines
 *	 char *  s2_currtime( )
 *	 int s2_usleep( microsecond )
 *	 int s2_sleep( second )
 *	 
 *	If option is 1, error message is printed on to stderr 
 *	when error occured.
 *	If option is 3, error and information message is printed on to stderr 
 *	when error occured.
 *	Error code errno returned from the system subroutines are preserved 
 *	on return to the caller.
 *	For usage, see comsubs2.doc or example programs such as
 *	s2_recho, s2_becho, s2_oecho, s2_xfer, etc.
 */

static char comsub_what[] = 
"@(#)comsub.c v1.12.3 tcp|udp client routines, K.Furukawa, Nov.1988-Dec.2006"; /*
**/

#include "comsub.h"			/* comsub definitions */
#include "comsub_p.h"			/* comsub private definitions */

/*static*/ struct  sockaddr_in  *s2_bdctin;	/* socket address structure */

static void s2_null( void );

#ifdef WINSOCK
void s2_initwinsock()
{
    static initwinsock = 1;
    if (initwinsock) {
         WSADATA wsadata;
         WSAStartup(MAKEWORD(1,1),&wsadata);
         initwinsock = 0;
    }
}
#endif /* WINSOCK */

/**
 *  s2_open
 *	s2_open prepares communication socket
 *
 *    int s2_open( portnum, service, protocol, hostname, option )
 *
 *	if portnum is positive, it tries to open a socket to hostname 
 *	with the portnum and protocol,
 *	then tries to open with service (service name or port number string)
 *	then if portnum is negative, tries to open with -portnum
 *	protocol is "tcp" or "udp"
 *	hostname is a host name or ip address in dotted notation
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than tcp/udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_open( 
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  char *protocol,		/* "tcp" or "udp" */
  char *hostname,		/* host name or ip-address */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_open";			/* name of the routine */
    int sd;					/* socket descriptor */
    struct  sockaddr_in  scin;			/* socket for connection */
    int usetcp;					/* tcp or udp */

#ifdef WINSOCK
    s2_initwinsock();
#endif /* WINSOCK */

/* tcp or udp */

    if ( ( !strcmp(protocol, "tcp") ) || ( !strcmp(protocol, "TCP") ) )  {
	usetcp = 1;				/* tcp protocol specified */
    }  else  {
	if ( ( !strcmp(protocol, "udp") ) || ( !strcmp(protocol, "UDP") ) )  {
	    usetcp = 0;				/* udp protocol specified */
	}  else  {
	    if ( option & S2_OALL )
		s2_errmsg( 0, nm, "protocol other than tcp/udp specified" );
	    return S2_NOPROT;			/* bad protocol name */
	}
    }

/* open socket and connect */

    sd = s2_preopen( &scin, portnum, service, usetcp, hostname, option );

    if ( sd < 0 )  return sd;			/* error return */

    sd = s2_postopen( &scin, sd, option );

    if ( sd < 0 )  return sd;			/* error return */

/* information */

    if ( option & S2_OINFO )  {
	static  struct  sockaddr_in  to;	/* socket for communication */
	int	tolen;

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
						/* get socket info. */
	    s2_errmsg( 1, nm, "warning - getsockname failure" );
	}
	fprintf( stderr, " Connected to %s on local port %u at %s\n", 
		hostname, ntohs( to.sin_port ), s2_currtime() );
    }
/* successful return */

    return sd;					/* socket descriptor */
}


/**
 *  s2_bopen
 *	s2_bopen prepares udp communication socket for broadcasting
 *
 *    int s2_bopen( portnum, service, protocol, hostname, option )
 *
 *	if portnum is positive, it tries to open a socket to hostname 
 *	with the portnum
 *	then tries to open with service (service name or port number string)
 *	then if portnum is negative, tries to open with -portnum
 *	only "udp" protocol is meaningful for broadcast
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *	if protocol is not "udp", S2_NOPROT is returned
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure/ broadcast assigning failure
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_bopen( 
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  char *protocol,		/* "udp" */
  char *hostname,		/* host name or ip-address */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_bopen";			/* name of the routine */
    int sd;					/* socket descriptor */
    static struct  sockaddr_in  bscin;		/* socket for connection */
    int usetcp;					/* tcp or udp */

/* tcp or udp */

    if ( strcmp(protocol, "udp") && strcmp(protocol, "UDP") )  {
	if ( option & S2_OALL )			/* if not "udp" */
	    s2_errmsg( 0, nm, 
		      "protocol other than udp specified for broadcast" );
	return S2_NOPROT;			/* bad protocol name */
    }
    usetcp = 0;				/* udp protocol specified */

/* open socket */

    sd = s2_preopen( &bscin, portnum, service, usetcp, hostname, option );

    if ( sd < 0 )  return sd;			/* error return */

/* broadcast socket option */

    {
	int on = 1;				/* set broadcast attribure */
	if (setsockopt(sd,SOL_SOCKET,SO_BROADCAST,(char *)&on,sizeof on)==-1) {
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "setsockopt(broadcast) failure" );
	    return S2_UBNDERR;			/* udp failure */
	}
    }

    s2_bdctin = &bscin;				/* save info for s2_bsend */
    /* in order to receive messages from many servers, we cannot
       connect socket to broadcast address.  if the connection is
       one-way like xfer program, we can connect.  otherwise we
       should save destination info. for s2_bsend. */

/* information */

    if ( option & S2_OINFO )  {
	static  struct  sockaddr_in  to;	/* socket for communication */
	int	tolen;

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
	    					/* get socket info. */
	    s2_errmsg( 1, nm, "warning - getsockname failure" );
	}
	fprintf( stderr, " Broadcast to %s on local port %u at %s\n", 
		hostname, ntohs( to.sin_port ), s2_currtime() );
    }

/* successful return */

    return sd;					/* socket descriptor */
}


/**
 *  s2_scopen
 *	s2_scopen prepares communication socket (sc_open compatible)
 *
 *    int s2_scopen( destination, protocol, option )
 *
 *	destination is a string, one of these forms
 *	 "service_name", "port_number",
 *	 "service_name@host_name", "port_number@host_name"
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than tcp/udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_scopen( 
  char *destination,		/* service and host */
  char *protocol,		/* "tcp" or "udp" */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_scopen";			/* name of the routine */
    char service[20];				/* service name */
    char *servicep = service;
    int servicel;
    char host[20];				/* host name */
    char *hostp = "127.0.0.1";
    char *point;

    servicel = strlen(destination);
    if ( (point = (char *)strchr(destination, '@')) )  {
	servicel = point-destination;
	if ( destination[servicel+1] != '\0' )  {
	    strncpy( host, destination+servicel+1, 20 );
	    hostp = host;
	}
    }
    strncpy( service, destination, servicel );
    service[servicel] = '\0';
    if ( service[0] == '#' )
	servicep++;

    return s2_open( 0, servicep, protocol, hostp, option );
}


/*
 *  s2_preopen
 *	s2_preopen prepares information for communication socket
 *	(internal routine for s2_open and s2_bopen)
 *
 *    int s2_preopen( scinp, portnum, service, usetcp, hostname, option )
 *
 *	scin is the pointer to socket structure.
 *	if portnum is positive, try to open a socket to hostname 
 *	with the protnum and protocol,
 *	then try to open with service
 *	then if portnum is negative, try to open with -portnum
 *	usetcp is 1 for tcp, 0 for udp
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure
 */
int  s2_preopen( 
  struct sockaddr_in *scinp,	/* socket structure for connection */
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  int usetcp,			/* protocol 1: tcp, 0: udp */
  char *hostname,		/* host name or ip-address */
  int option			/* option 1: error, 2: info. message */
)
/*
*/
{
    char *nm = "s2_preopen";			/* name of the routine */
    int sd;					/* socket descriptor */
#ifndef NONETDB
    struct  servent  *sp;			/* service data structure */
    struct  hostent  *hp;			/* host data structure */
#endif /* NONETDB */

    memset( (char *)scinp, 0, sizeof(struct sockaddr_in) );
						/* initialize data structure */

/* port/service handling */

    if ( portnum > 0 )  {			/* port number specified */
	scinp->sin_port = htons( portnum );	/* set service port number */
    }  else if ( service && strlen( service ) )  {
#ifndef NONETDB
	sp = getservbyname( service, (usetcp ? "tcp" : "udp" ) );
	if ( sp )  {				/* get service information */
	    scinp->sin_port = sp->s_port;  	/* set port number */
	}  else	
#endif /* NONETDB */
	{
	    int port;
	    if ( *service == '#' )  service++;
	    if ( (port = atoi(service)) > 0 )  {
		scinp->sin_port = htons( port );  /* port number in string */
	    }  else if ( portnum < 0 )  {
		scinp->sin_port = htons( -portnum );  /* default port */
	    }  else  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "service not found\n" );
		return S2_NOSERV;		/* service not found */
	    }
	}
    }  else if ( portnum < 0 )  {
	scinp->sin_port = htons( -portnum );	/* default port */
    }  else  {
	if ( option & S2_OALL )
	    s2_errmsg( 0, nm, "no port/service specified\n" );
	return S2_NOPORT;			/* no port/service specified */
    }

/* host handling */

    if ( !hostname || !*hostname )  {		/* null pointer/null string */
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "unknown host" );
	return S2_NOHOST;			/* unknown host */
    }

#ifndef NONETDB
    hp = gethostbyname( hostname );		/* get host information */
    if ( hp )  {
	memcpy( (char *)&(scinp->sin_addr), hp->h_addr, hp->h_length );
	scinp->sin_family = hp->h_addrtype;  	/* set host address/type */
    }  else  
#endif /* NONETDB */
    {
	LInt32 adr;				/* dot formatted address */
	adr = (LInt32) inet_addr( hostname );
	if ( adr != -1 )  {			/* get address in dot format */
	    memcpy( (char *)&(scinp->sin_addr), &adr, sizeof( adr ) );
	    scinp->sin_family = AF_INET;
	}  else  {
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "unknown host" );
	    return S2_NOHOST;			/* unknown host */
	}
    }

/* socket handling */

    sd = socket( AF_INET, ( usetcp ? SOCK_STREAM : SOCK_DGRAM ), 0 );
						/* create connection socket */
    if ( sd == -1 )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "socket open failure" );
	return S2_OPENERR;			/* socket open failure */
    }

    if ( !usetcp )  {				/* for udp */
	static struct sockaddr_in myin;		/* socket for local */

	myin.sin_family = AF_INET;
	myin.sin_port = 0;
	myin.sin_addr.s_addr = INADDR_ANY;
	if ( -1 == bind(sd, (void *)&myin, sizeof(myin)) )  {
						/* bind local address */
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "udp bind failure" );
	    return S2_UBNDERR;			/* udp bind failure */
	}
    }

/* successful return */

    return sd;					/* socket descriptor */
}

/*
 *  s2_postopen
 *	s2_postopen connects socket prepared by s2_preopen
 *	(internal routine for s2_open and s2_bopen)
 *
 *    int s2_postopen( scinp, sd, option )
 *
 *	scinp is the pointer to socket structure.
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_postopen( 
  struct sockaddr_in *scinp,	/* socket structure for connection */
  int sd,			/* socket descript. returned from s2_preopen */
  int option			/* option 1: error, 2: info. message */
)
/*
*/
{
    char *nm = "s2_postopen";			/* name of the routine */

/* connect */

    if ( -1 == connect(sd, (void *)scinp, sizeof(struct sockaddr_in)) )  {
	if ( option & S2_OALL )			/* connect */
	    s2_errmsg( 1, nm, "socket connect failure" );
	close( sd );
	return S2_CONNERR;			/* socket connect failure */
    }

/* successful return */

    return sd;					/* socket descriptor */
}


/**
 *  s2_close
 *	s2_close close communication port
 *
 *    int  s2_close( sd, option )
 *
 *    return:	always zero
 */
int  s2_close( 
  int sd,			/* socket descriptor returned from s2_open */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_close";			/* name of the routine */

/* information */

    if ( option & S2_OINFO )  {
	static  struct  sockaddr_in  to;	/* socket for communication */
	int	tolen;

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
						/* get socket info. */
	    s2_errmsg( 1, nm, "warning - getsockname failure" );
	}
	fprintf( stderr, 
		" Disconnected local port %u at %s\n", 
		ntohs( to.sin_port ), s2_currtime() );
    }

    shutdown( sd, 2 );				/* close communication */
#ifndef vms
#ifndef WINSOCK
    close( sd );				/* close socket */
#else /* WINSOCK */
    closesocket( sd );				/* close socket */
#endif /* WINSOCK */
#else /* vms */
    netclose( sd );
#endif /* vms */
    return 0;
}


/**
 *  s2_send
 *	s2_send sends message to communication port
 *	calls send(2) system call.
 *
 *    int  s2_send( sd, buffer, buflen, option )
 *
 *    return:	-21 S2_SENDERR	send failure
 *		positive	byte count which has been sent
 */
int  s2_send( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to send */
  int buflen,			/* buffer length */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_send";			/* name of the routine */
    int iret;

    if (buflen > (iret = send(sd, buffer, buflen, 0) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "send failure" );
	return S2_SENDERR;			/* send failure */
    }
    return iret;
}


/**
 *  s2_receive
 *	s2_receive receives message from communication port
 *	calls recv(2) system call.
 *
 *    int  s2_receive( sd, buffer, buflen, option )
 *
 *    return:	-22 S2_RECVERR	receive failure
 *		positive	byte count which has been received
 */
int  s2_receive( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to receive */
  int buflen,			/* buffer length */
  int option			/* option 1: error/information message */
)
/*
**/
{
    char *nm = "s2_receive";			/* name of the routine */
    int iret;

    if (-1 == (iret = recv(sd, buffer, buflen, 0) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "recv failure" );
	return S2_RECVERR;			/* receive failure */
    }
    return iret;
}


/**
 *  s2_bsend
 *	s2_bsend sends message to broadcast address
 *	calls sendto(2) system call.  we need separate send routine
 *	for bi-directional broadcast-and-reply communication
 *	this routine should be used after s2_bopen
 *
 *    int  s2_bsend( sd, buffer, buflen, option )
 *
 *    return:	-21 S2_SENDERR	send failure
 *		positive	byte count which has been sent
 */
int  s2_bsend( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to send */
  int buflen,			/* buffer length */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_bsend";			/* name of the routine */
    int iret;
    int tolen;

    tolen = sizeof (struct  sockaddr_in);

    if (buflen > 
	(iret = sendto(sd, buffer, buflen, 0, (void *)s2_bdctin, tolen) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "send failure" );
	return S2_SENDERR;			/* send failure */
    }
    return iret;
}


/**
 *  s2_currtime1
 *	returns current time in second and micro-second
 */
int s2_currtime1(
  time_t *second,		/* current time "second" part */
  int *micro			/* current time "microsecond" part */
)
/*
**/
{

#ifdef USETIME				/* use time() */
    time_t timet;

    time( &timet );
    *second = timet;
    *micro = 0;
#else
#ifdef USEFTIME				/* use ftime() */
    struct timeb timebt;

    ftime( &timebt );
    *second = timebt.time;
    *micro = (int)timebt.millitm * 1000;
#else					/* use gettimeofday */
    struct timeval timev;
    /* struct timezone is not standard anymore at least on tru64, linux, ... */
    /* struct timezone timez; */

    /* gettimeofday( &timev, &timez ); */
    gettimeofday( &timev, NULL );
    *second = timev.tv_sec;
    *micro = (int)timev.tv_usec;
#endif
#endif
    return (int)*second;
}

/**
 *  s2_currtime
 *	returns string of current time
 *	24 characters like "day mmm dd hh:mm:ss yyyy"
 *	or 27 characters like "day mmm dd hh:mm:ss.sss yyyy"
 */
char *s2_currtime( void )
/*
**/
{
    char *c;
    static char chartime[30];
    time_t second;
    int micro;

    (void) s2_currtime1( &second, &micro );
    c = ctime( &second );			/* convert into string */
    if ( micro )  {			/* day mmm dd hh:mm:ss.sss yyyy */
	memcpy( chartime, c, 19 );
	sprintf( chartime+19, ".%03d ", micro/1000 );
	memcpy( chartime+24, c+20, 4 );
	chartime[28] = '\0';
    }  else  {				/* day mmm dd hh:mm:ss yyyy */
	memcpy( chartime, c, 24 );
	chartime[24] = '\0';
    }
    return chartime;
}

/**
 *  s2_usleep
 *	s2_usleep is equivalent to usleep on many unix systems. 
 *	it was mainly prepared for ultrix compatibility.
 *
 *    int  s2_usleep(micro)
 *
 *    return:	-23 S2_SLCTERR	select failure
 *		0		byte count which has been sent
 */
int  s2_usleep( 
  int micro			/* micro second to sleep */
)
/*
**/
{
    char *nm = "s2_usleep";			/* name of the routine */

#ifdef NANOSLEEP
    struct timespec ts;
    ts.tv_sec = micro / 1000000L;
    ts.tv_nsec = (micro % 1000000L) * 1000;
    nanosleep( ts, NULL );
#else /* NANOSLEEP */
#ifndef NOSELECT
    struct timeval tv;
    tv.tv_sec = micro / 1000000;
    tv.tv_usec = micro % 1000000;

    if (select (1, 0, 0, 0, &tv) < 0)  return S2_SLCTERR;
#else /* NOSELECT */
#ifdef USLEEP
    if ( usleep( micro ) )  return S2_SLCTERR;
#else /* USLEEP */
    sleep(micro / 1000000);
#endif /* USLEEP */
#endif /* NOSELECT */
#endif /* NANOSLEEP */
    return 0;
}

/**
 *  s2_sleep
 *	s2_sleep is equivalent to sleep on many unix systems. 
 *	it was mainly prepared for vxworks compatibility.
 *
 *    int  s2_sleep(second)
 *
 *    return:	-23 S2_SLCTERR	select failure
 *		0		byte count which has been sent
 */
int  s2_sleep( 
  int second			/* second to sleep */
)
/*
**/
{
    char *nm = "s2_sleep";			/* name of the routine */

    return s2_usleep( second*1000000 );
}

/*
 *  s2_errmsg
 *	print error message
 */
void
 s2_errmsg( flg, nm, msg )
int flg;				/* flag to use perror */
char * nm;				/* routine name where error occur*/
char * msg;				/* error message to display */
{
    /* extern int errno; */
    int errtmp;
    errtmp = errno;
    if ( flg )  perror( nm );
    fprintf( stderr, "%s: %s\n", nm, msg );
    errno = errtmp;
}

static void s2_null( void )
{
    return;
}

