/**
 *	comsubt  --  communication subroutines with timeout
 *		Copyright (C) Kazuro Furukawa, 1988-2004
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa  Nov.9.1988.  (rwrite)
 *      K.F  Apr.13.1989.  converted from rwrite (recho)
 *	K.F  Mar.19.1991.  recho v3.4.2
 *	K.F  Mar.24.1992.  comsub
 *	K.F  Jun.12.1993.  comsubt
 *	Modifications
 *	K.F  Jun.12.1993.  add s2_tureceive() and s2_tusend()
 *	K.F  Aug.29.1995.  add s2_topen()
 *	K.F  Feb.21.1996.  ansi C
 *	K.F  May.18.1999.  move s2_usleep to comsub.c
 *
 *	This module defines send and receive rouine for socket communi-
 *	cation.  It can be used with timeout specification.  If you 
 *	don't need timeout function, you may use s2_open, s2_send and 
 *	s2_receive.  Timeout value is specified in milli-seconds.
 *	 int s2_topen( portnum, service, protocol, hostname, timeout, 
 *	              option )
 *	 int s2_tsend( sd, buffer, buflen, timeout, option )
 *	 int s2_treceive( sd, buffer, buflen, timeout, option )
 *	For udp server, you should use these functions instead.  If you 
 *	don't need timeout function, you may use s2_usend and s2_ureceive.
 *	 int s2_tusend( sd, buffer, buflen, timeout, option )
 *	 int s2_tureceive( sd, buffer, buflen, timeout, option )
 *	If option is S2_OERR, error message is printed on to stderr 
 *	when error occured.
 *	If option is S2_OALL, error and information message is printed 
 *	on to stderr when error occured.
 *	Error codes errno returned from the system subroutines are 
 *	preserved on return to the caller.
 *
 *	Bugs:
 *	 not all environments are supported.  ex. MSDOS is not supported.
 */

char	comsubt_what[] = 
"@(#)comsubt.c v1.3.6 tcp|udp client subroutines, K.Furukawa, Feb.1996-Sep.2000"; /*
**/

#include "comsub.h"			/* comsub definitions */
#include "comsub_p.h"			/* comsub private definitions */

#include <errno.h>
#ifndef NOSELECT
#ifndef NOSOCKFCNTL
#include <fcntl.h>
#endif /* NOSOCKFCNTL */
#endif /* NOSELECT */

/**
 *  s2_topen
 *	s2_topen prepares communication socket with timeout
 *
 *    int s2_topen(portnum, service, protocol, hostname, timeout, option)
 *
 *	if portnum is positive, it tries to open a socket to hostname 
 *	with the portnum and protocol,
 *	then tries to open with service (service name or port number string)
 *	then if portnum is negative, tries to open with -portnum
 *	protocol is "tcp" or "udp"
 *	hostname is a host name or ip address in dotted notation
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than tcp/udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure
 *		-7 S2_CONNERR	socket connect failure
 *    		-23 S2_SLCTERR	select or fcntl failure
 *    		-24 S2_TIMEOUT	timeout failure
 */
int  s2_topen(
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  char *protocol,		/* "tcp" or "udp" */
  char *hostname,		/* host name or ip-address */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_topen";			/* name of the routine */
    int s2_tpostopen( struct sockaddr_in *, int, int, int );
    int sd;					/* socket descriptor */
    struct  sockaddr_in  scin;			/* socket for connection */
    int usetcp;					/* tcp or udp */

/* tcp or udp */

    if ( ( !strcmp(protocol, "tcp") ) || ( !strcmp(protocol, "TCP") ) )  {
	usetcp = 1;				/* tcp protocol specified */
    }  else  {
	if ( ( !strcmp(protocol, "udp") ) || ( !strcmp(protocol, "UDP") ) )  {
	    usetcp = 0;				/* udp protocol specified */
	}  else  {
	    if ( option & S2_OALL )
		s2_errmsg( 0, nm, "protocol other than tcp/udp specified" );
	    return S2_NOPROT;			/* bad protocol name */
	}
    }

/* open socket and connect */

    sd = s2_preopen( &scin, portnum, service, usetcp, hostname, option );

    if ( sd < 0 )  return sd;			/* error return */

    sd = s2_tpostopen( &scin, sd, timeout, option );

    if ( sd < 0 )  return sd;			/* error return */

/* information */

    if ( option & S2_OINFO )  {
	static  struct  sockaddr_in  to;	/* socket for information */
	int	tolen;
	char	*s2_currtime();			/* current time */

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
						/* get socket info. */
	    s2_errmsg( 1, nm, "warning - getsockname failure" );
	}
	fprintf( stderr, " Connected to %s on local port %u at %s\n", 
		hostname, ntohs( to.sin_port ), s2_currtime() );
    }
/* successful return */

    return sd;					/* socket descriptor */
}


/*
 *  s2_tpostopen
 *	s2_tpostopen connects socket prepared by s2_preopen with timeout
 *	(internal routine for s2_topen)
 *
 *    int s2_tpostopen(scinp, sd, timeout, option)
 *
 *	scinp is the pointer to socket structure.
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-7 S2_CONNERR	socket connect failure
 *    		-23 S2_SLCTERR	select or fcntl failure
 *    		-24 S2_TIMEOUT	timeout failure
 */
int  s2_tpostopen( 
  struct sockaddr_in *scinp,	/* socket structure for connection */
  int sd,			/* socket descript. returned from s2_preopen */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1: error, 2: info. message */
)
/*
*/
{
    char *nm = "s2_tpostopen";			/* name of the routine */

#if defined(NOSELECT) || defined(NOSOCKFCNTL)
    if ( option & S2_OINFO )
	s2_errmsg( 0, nm, "no timeout (fcntl) supported on this machine" );
#else
    struct timeval timeoutval;
    fd_set writes;
    int iret;

    timeoutval.tv_sec = timeout/1000;
    timeoutval.tv_usec = (timeout%1000)*1000;

    iret = fcntl(sd, F_SETFL, O_NDELAY);
    if (iret == -1)  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "fcntl failure" );
	close( sd );
	return S2_SLCTERR;			/* fcntl failure */
    }
#endif /* NOSELECT, NOSOCKFCNTL */	    

/* connect */

    if ( -1 == connect(sd, (void *)scinp, sizeof(struct sockaddr_in)) )  {
						/* connect */
	if ( errno != EWOULDBLOCK && errno != EINPROGRESS)  {
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "socket connect failure" );
	    close( sd );
	    return S2_CONNERR;			/* socket connect failure */
	}
    }

#if !defined(NOSELECT) && !defined(NOSOCKFCNTL)
    do  {
	FD_ZERO(&writes);
	FD_SET(sd, &writes);
	iret = select(sd+1, (fd_set_p)0, (fd_set_p)&writes, (fd_set_p)0, 
		      &timeoutval);
	if (!iret)  {
	    if ( option & S2_OINFO )
		s2_errmsg( 1, nm, "connect timeout" );
	    close( sd );
	    return S2_TIMEOUT;			/* timeout */
	}
	if (iret < 0)  {
	    if ( errno != EINTR )  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "select failure" );
		close( sd );
		return S2_SLCTERR;		/* select failure */
	    }
	}
    } while(iret != 1);				/* until socket is ready */

    /* here, select returned without timeout, which means connect */
    /* finished.  however, we don't know whether connect succeeds */
    /* or is refused.  try to check it with getsockname.          */
    /* i don't know if this is the correct way or not.            */

    {
	struct  sockaddr_in  to;		/* socket for information */
	int	tolen;

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
						/* get socket info. */
	    if ( option & S2_OALL )
		s2_errmsg( 1, nm, "socket connect refused" );
	    close( sd );
	    return S2_CONNERR;			/* socket connect failure */
	}
    }

#endif /* NOSELECT, NOSOCKFCNTL */	    

/* successful return */

    return sd;					/* socket descriptor */
}


/**
 *  s2_tsend
 *	s2_tsend sends message to communication port with timeout
 *	calls send(2) system call.
 *
 *    int  s2_tsend(sd, buffer, buflen, timeout, option)
 *
 *    return:	-21 S2_SENDERR	send failure
 *    		-23 S2_SLCTERR	select failure
 *    		-24 S2_TIMEOUT	timeout failure
 *		positive	byte count which has been sent
 */
int  s2_tsend( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to send */
  int buflen,			/* buffer length */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_tsend";			/* name of the routine */
    int iret;

#ifdef NOSELECT
    if ( option & S2_OINFO )
	s2_errmsg( 0, nm, "no timeout (select) supported on this machine" );
#else
    struct timeval timeoutval;
    fd_set writes;

    timeoutval.tv_sec = timeout/1000;
    timeoutval.tv_usec = (timeout%1000)*1000;

    do  {
	FD_ZERO(&writes);
	FD_SET(sd, &writes);
	iret = select(sd+1, (fd_set_p)0, (fd_set_p)&writes, (fd_set_p)0, 
		      &timeoutval);
	if (!iret)  {
	    if ( option & S2_OINFO )
		s2_errmsg( 0, nm, "send timeout" );
	    return S2_TIMEOUT;			/* timeout */
	}
	if (iret < 0)  {
	    if ( errno != EINTR )  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "select failure" );
		return S2_SLCTERR;		/* select failure */
	    }
	}
    } while(iret != 1);				/* until socket is ready */
#endif /* NOSELECT */	    
	    
    if (buflen > (iret = send(sd, buffer, buflen, 0) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "send failure" );
	return S2_SENDERR;			/* send failure */
    }
    return iret;
}


/**
 *  s2_treceive
 *	s2_treceive receives message from communication port with timeout
 *	calls recv(2) system call.
 *
 *    int  s2_treceive(sd, buffer, buflen, timeout, option)
 *
 *    return:	-22 S2_RECVERR	receive failure
 *    		-23 S2_SLCTERR	select failure
 *    		-24 S2_TIMEOUT	timeout failure
 *		positive	byte count which has been received
 */
int  s2_treceive( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to receive */
  int buflen,			/* buffer length */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_treceive";			/* name of the routine */
    int iret;

#ifdef NOSELECT
    if ( option & S2_OINFO )
	s2_errmsg( 0, nm, "no timeout (select) supported on this machine" );
#else
    struct timeval timeoutval;
    fd_set reads;

    timeoutval.tv_sec = timeout/1000;
    timeoutval.tv_usec = (timeout%1000)*1000;

    do  {
	FD_ZERO(&reads);
	FD_SET(sd, &reads);
	iret = select(sd+1, (fd_set_p)&reads, (fd_set_p)0, (fd_set_p)0, 
		      &timeoutval);
	if (!iret)  {
	    if ( option & S2_OINFO )
		s2_errmsg( 0, nm, "receive timeout" );
	    return S2_TIMEOUT;			/* timeout */
	}
	if (iret < 0)  {
	    if ( errno != EINTR )  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "select failure" );
		return S2_SLCTERR;		/* select failure */
	    }
	}
    } while(iret != 1);				/* until socket is ready */
#endif /* NOSELECT */	    

    if (-1 == (iret = recv(sd, buffer, buflen, 0) ) )  {
	if ( option & S2_OALL )
	    s2_errmsg( 1, nm, "recv failure" );
	return S2_RECVERR;			/* receive failure */
    }
    return iret;
}


/**
 *  s2_tureceive
 *	s2_tureceive receives message from "udp" communication port with
 *	timeout,  calls recvfrom(2) system call.
 *
 *    int  s2_tureceive(sl, buffer, buflen, timeout, option)
 *
 *    return:	-22 S2_RECVERR	receive failure
 *		-23 S2_SLCTERR	select failure
 *		-24 S2_TIMEOUT	timeout failure
 *		positive	byte count which has been received
 */
int  s2_tureceive( 
  int sd,			/* socket descriptor returned from s2_sopen */
  char *buffer,			/* buffer to receive */
  int buflen,			/* buffer length */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1: error/information message */
)
/*
**/
{
    char *nm = "s2_tureceive";			/* name of the routine */
    int iret;

#ifdef NOSELECT
    if ( option & S2_OINFO )
	s2_errmsg( 0, nm, "no timeout (select) supported on this machine" );
#else
    struct timeval timeoutval;
    fd_set reads;

    timeoutval.tv_sec = timeout/1000;
    timeoutval.tv_usec = (timeout%1000)*1000;

    do  {
	FD_ZERO(&reads);
	FD_SET(sd, &reads);
	iret = select(sd+1, (fd_set_p)&reads, (fd_set_p)0, (fd_set_p)0, 
		      &timeoutval);
	if (!iret)  {
	    if ( option & S2_OINFO )
		s2_errmsg( 0, nm, "receive timeout" );
	    return S2_TIMEOUT;			/* timeout */
	}
	if (iret < 0)  {
	    if ( errno != EINTR )  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "select failure" );
		return S2_SLCTERR;		/* select failure */
	    }
	}
    } while(iret != 1);				/* until socket is ready */
#endif /* NOSELECT */	    

    return s2_ureceive(sd, buffer, buflen, option);
}


/**
 *  s2_tusend
 *	s2_tusend sends message to "udp" communication port with timeout
 *	calls sendto(2) system call.
 *
 *    int  s2_tusend(sl, buffer, buflen, timeout, option)
 *
 *    return:	-21 S2_SENDERR	send failure
 *		-23 S2_SLCTERR	select failure
 *		-24 S2_TIMEOUT	timeout failure
 *		positive	byte count which has been sent
 */
int  s2_tusend( 
  int sd,			/* socket descriptor returned from s2_open */
  char *buffer,			/* buffer to send */
  int buflen,			/* buffer length */
  int timeout,			/* timeout in milli-second */
  int option			/* option 1/2: error/information message */
)
/*
**/
{
    char *nm = "s2_tusend";			/* name of the routine */
    int iret;

#ifdef NOSELECT
    if ( option & S2_OINFO )
	s2_errmsg( 0, nm, "no timeout (select) supported on this machine" );
#else
    struct timeval timeoutval;
    fd_set writes;

    timeoutval.tv_sec = timeout/1000;
    timeoutval.tv_usec = (timeout%1000)*1000;

    do  {
	FD_ZERO(&writes);
	FD_SET(sd, &writes);
	iret = select(sd+1, (fd_set_p)0, (fd_set_p)&writes, (fd_set_p)0, 
		      &timeoutval);
	if (!iret)  {
	    if ( option & S2_OINFO )
		s2_errmsg( 0, nm, "send timeout" );
	    return S2_TIMEOUT;			/* timeout */
	}
	if (iret < 0)  {
	    if ( errno != EINTR )  {
		if ( option & S2_OALL )
		    s2_errmsg( 1, nm, "select failure" );
		return S2_SLCTERR;		/* select failure */
	    }
	}
    } while(iret != 1);				/* until socket is ready */
#endif /* NOSELECT */	    

    return s2_usend(sd, buffer, buflen, option);
}

