/**
 *	comsubu  --  communication subroutines for udp without connect
 *		Copyright (C) Kazuro Furukawa, 1988-2004
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *		(if you modified this code, please inform furukawa)
 *
 *	K.Furukawa  Nov.9.1988.  (rwrite)
 *      K.F  Apr.13.1989.  converted from rwrite (recho)
 *	K.F  Mar.19.1991.  recho v3.4.2
 *	K.F  Mar.24.1992.(2)  comsub
 *	K.F  Sep.14.1992.  add option 3 (bit 2)
 *	K.F  Mar.4.1994.   udp without connection
 *	Modifications
 *	K.F  Feb.21.1996.  ansi C
 *
 *	This module is a re-writing of zechosub.
 *	This module defines basic rouines for udp socket communication.
 *	 int s2_uopen( portnum, service, protocol, hostname, option )
 *	 int s2_scuopen( destination, protocol, option )
 *
 *	You may use only one socket with uopen routines.
 *	s2_bsend and s2_breceive routine can be used with these routines
 *
 *	If option is 1, error message is printed on to stderr 
 *	when error occured.
 *	If option is 3, error and information message is printed on to stderr 
 *	when error occured.
 *	Error code errno returned from the system subroutines are preserved 
 *	on return to the caller.
 *	For usage, see comsubs2.doc or example programs such as
 *	s2_in2soc, etc.
 */

char	comsubu_what[] = 
"@(#)comsubu.c v1.1.2 udp client subroutines, K.Furukawa, Feb.1996"; /*
**/

#include "comsub.h"			/* comsub definitions */
#include "comsub_p.h"			/* comsub private definitions */

/**
 *  s2_uopen
 *	s2_uopen prepares udp communication socket without connection
 *
 *    int s2_uopen( portnum, service, protocol, hostname, option )
 *
 *	this is based on s2_bopen
 *	if portnum is positive, try to open a socket to hostname 
 *	with the portnum
 *	then try to open with service
 *	then if portnum is negative, try to open with -portnum
 *	only "udp" protocol is meaningful for this routine
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *	if protocol is not "udp", S2_NOPROT is returned
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure/ broadcast assigning failure
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_uopen( 
  int portnum,			/* positive: port no, negative: default port */
  char *service,		/* service name, or null */
  char *protocol,		/* "udp" */
  char *hostname,		/* host name or ip-address */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_uopen";			/* name of the routine */
    int sd;					/* socket descriptor */
    static struct  sockaddr_in  bscin;		/* socket for connection */
    int usetcp;					/* tcp or udp */
    extern struct sockaddr_in *s2_bdctin;	/* save info for s2_bsend */

/* tcp or udp */

    if ( strcmp(protocol, "udp") && strcmp(protocol, "UDP") )  {
	if ( option & S2_OALL )			/* if not "udp" */
	    s2_errmsg( 0, nm, 
		      "protocol other than udp specified for broadcast" );
	return S2_NOPROT;			/* bad protocol name */
    }
    usetcp = 0;				/* udp protocol specified */

/* open socket */

    sd = s2_preopen( &bscin, portnum, service, usetcp, hostname, option );

    if ( sd < 0 )  return sd;			/* error return */

/* no socket option */

    s2_bdctin = &bscin;				/* save info for s2_bsend */
    /* in order to live with server resets, we cannot
       connect socket to destination address.  
       we should save destination info. for s2_bsend. */

/* information */

    if ( option & S2_OINFO )  {
	static  struct  sockaddr_in  to;	/* socket for communication */
	int	tolen;
	char	*s2_currtime();			/* current time */

	tolen = sizeof( to );
	if ( getsockname( sd, (void *)&to, &tolen ) == -1 ) {
						/* get socket info. */
	    s2_errmsg( 1, nm, "warning - getsockname failure" );
	}
	fprintf( stderr, " Opened to %s on local port %u at %s\n", 
		hostname, ntohs( to.sin_port ), s2_currtime() );
    }

/* successful return */

    return sd;					/* socket descriptor */
}


/**
 *  s2_scuopen
 *	s2_scuopen prepares udp communication socket (sc_open compatible)
 *
 *    int s2_scuopen( destination, protocol, option )
 *
 *	destination is a string, one of these forms
 *	 "service_name", "#port_number",
 *	 "service_name@host_name", "#port_number@host_name"
 *	if option is 1 or S2_OERR, error message is sent to stderr
 *	if option is 3 or S2_OALL, error/info. message is sent to stderr
 *
 *    return:	positive/0	success, socket descriptor
 *		-1 S2_NOSERV	service not found
 *		-2 S2_NOPORT	no port/service specified
 *		-3 S2_NOHOST	unknown host
 *		-4 S2_NOPROT	protocol other than tcp/udp specified
 *		-5 S2_OPENERR	socket open failure
 *		-6 S2_UBNDERR	udp bind failure
 *		-7 S2_CONNERR	socket connect failure
 */
int  s2_scuopen( 
  char *destination,		/* service and host */
  char *protocol,		/* "tcp" or "udp" */
  int option			/* option 1: error, 2: info. message */
)
/*
**/
{
    char *nm = "s2_scuopen";			/* name of the routine */
    char service[20];				/* service name */
    char *servicep = service;
    int servicel;
    char host[20];				/* host name */
    char *hostp = "127.0.0.1";
    char *point;

    servicel = strlen(destination);
    if ( (point = (char *)strchr(destination, '@')) )  {
	servicel = point-destination;
	if ( destination[servicel+1] != '\0' )  {
	    strncpy( host, destination+servicel+1, 20 );
	    hostp = host;
	}
    }
    strncpy( service, destination, servicel );
    service[servicel] = '\0';
    if ( service[0] == '#' )
	servicep++;

    return s2_uopen( 0, servicep, protocol, hostp, option );
}
