/**
 *	s2_xferd  --  message transfer server program
 *		Copyright (C) Kazuro Furukawa, 1988-1993
 *		Kazuro Furukawa, KEK, Japan <kazuro.furukawa@kek.jp>
 *
 *	K.Furukawa  Nov.9.1988. (rwrited)
 *	K.F  Apr.13.1989.  converted from rwrited to rechod
 *      K.F  Mar.12.1991.  converted from rechod to xferd
 *	K.F  Jun.17.1992.  rechod v3.4.2
 *	K.F  Oct.2.1992.   s2_rechod using s2_xxx routines
 *	K.F  Nov.26.1992.  s2_rechod v1.2.0
 *	K.F  Jun.10.1993.  s2_xferd using s2_xxx routines
 *	Modifications
 *	K.F  Feb.21.1996.  ansi C
 *
 *      This program receives messages from s2_xfer program and outputs
 *	them to stdout.
 *	Xfer and xferd process one-way transmittion of messages while
 *	recho and rechod communicate in bi-directional way.
 *	With tcp, xfer can send all data from stdin to stdout of xferd.
 *	Thus xfer and xferd act as a pipe through the network.
 *	With udp, xfer may loose messages if the transfer rate is high.
 *	However xfer can send messages to many hosts with broadcast.
 *	This will be the typical example for the TCP|UDP/IP server-client
 *	model.
 *	See argument() for options.
**/

char	what[] = 
"@(#)s2_xferd.c v1.1.1 tcp|udp/ip echo server, K.Furukawa, Nov.1988-Aug.2003";

#include "comsub.h"				/* network header files */

#define  SERVICENAME	"s2_xfer"		/* service name */
#define  PORTNUMB	11062			/* default service port */
#define  BUFFERSIZE	16384			/* communication buffer size */

char	*nm;
char	buf[BUFFERSIZE];			/* communication buffer */
int	usetcp;					/* flag for protocol tcp/udp */
int	forkdaemon;				/* flag for daemon fork */
int	oneconnect;				/* flag for one connection */
int	portnum;				/* port number */

void errormsg( char * );

/**
 *  main
 *	Main prepares a listen socket for the xfer service and forks a daemon
 *	process in order to disconnect the control terminal.  
**/
int  main( int argc, char ** argv )
{
    void argument( int, char ** );
    void tcpserver( int );
    void udpserver( int );
    int	sl;					/* socket desctiptor */

    nm = argv[0];				/* program name */
    usetcp = 1;					/* use tcp by default */
    forkdaemon = 0;				/* no daemon by default */
    oneconnect = 0;				/* accept many connections */
    portnum = 0;				/* port number */
    
    argument(argc, argv);			/* command line argument */
    
    if ( !portnum )  {				/* port number not specified */
	portnum = -PORTNUMB;			/* default port */
    }

#ifndef SINGLESERVER
    if ( !forkdaemon )  {
#endif
	sl = s2_sopen( portnum, SERVICENAME, (usetcp?"tcp":"udp"), S2_OALL );
	if ( sl<0 )  exit( 1 );			/* create listen socket */

	if ( usetcp )  tcpserver( sl );		/* server for tcp */
	else  udpserver( sl );			/* server for udp */
	exit( 0 );

#ifndef SINGLESERVER
    } else {
	setpgrp();			/* disconnect control terminal */

	switch ( fork() )  {
	case -1:
	    errormsg( "daemon fork failure" );
	    exit( 1 );
	case 0:					/* daemon */
	    fclose( stdin );			/* close control terminal */
	    sl = s2_sopen(portnum, SERVICENAME, (usetcp?"tcp":"udp"), S2_OALL);
	    if ( sl<0 )  exit( 1 );		/* create listen socket */

	    if ( usetcp )  tcpserver( sl );	/* server for tcp */
	    else  udpserver( sl );		/* server for udp */

	default:				/* parent process */
	    exit( 0 );				/* free terminal */
	}
    }
#endif
    return 0;
}

/**
 *  tcpserver
 *	This routine waits for connections from clients.  If a request 
 *	arrives, it forks a child to process each request.
 *	Server reports the connection information to standard error (
 *	maybe logging file).  Then it sends the message received from 
 *	the client to stdout.  It also reports disconnection information.
**/
void  tcpserver( int sl )
{
    int	sn;					/* socket desctiptor */

    for(;;)  {
#ifndef SINGLESERVER
	sn = s2_accept( sl, 1, S2_OALL );	/* accept connection */
#else
	sn = s2_accept( sl, 0, S2_OALL );	/* accept connection */
#endif
	if (sn<0) exit(1);
	if (sn>0)  {
	    int	cc;
	    long lines = 0;			/* communication counter */
	    long bytes = 0;			/* communication counter */
						/* server main body starts */
	    while( ( cc = s2_receive(sn, buf, BUFFERSIZE, S2_OALL) ) > 0 )  {
						/* read message */
		if( write(1, buf, cc) < cc )  {	/* send to stdout */
		    errormsg( "write failure to stdout" );
		    exit( 1 );
		}
		lines++;
		bytes += cc;
	    }

	    fprintf( stderr, "  %ld lines %ld bytes\n", lines, bytes );
	    s2_sclose( sn, S2_OALL );
#ifndef SINGLESERVER
	    exit( 0 );				/* child exit */
#endif
	}
	if ( oneconnect )  exit( 0 );		/* -1 option, exit */
    }						/* parent repeat forever */
}

/**
 *  udpserver
 *	Udpserver waits for message from clients.  If a message arrives,
 *	it sends the message to stdout.
**/
void  udpserver( int sl )
{
    int	cc;

    while( ( cc = s2_ureceive( sl, buf, BUFFERSIZE, S2_OALL ) ) >= 0 )  {
						/* read message */
	if( write( 1, buf, cc ) < cc ) {	/* send to stdout */
	    errormsg( "write failure to stdout" );
	    exit( 1 );
	}

    }						/* repeat forever */
    exit( 1 );
}

/*
 *  errormsg
 *	print error message
 */
void  errormsg( char * msg )
{
	perror( nm );
	fprintf( stderr, "%s: %s\n", nm, msg );
}

/**
 *  argument
 *	Argument handles command line argument, such as hostname or options.
**/
void  argument( int argc, char ** argv )
{
    int c;
    extern int optind, opterr;
    extern char *optarg;
    opterr = 0;				/* disable informational message */

#ifndef SINGLESERVER
    while ((c = getopt (argc, argv, "dD1uUp:P:hH")) != EOF)  {
	switch ( c )  {
	case 'd':
	case 'D':				/* use udp */
	    forkdaemon = 1;			/* set fork daemon flag */
	    break;
#else
    while ((c = getopt (argc, argv, "1uUp:P:hH")) != EOF)  {
	switch ( c )  {
#endif
	case '1':				/* accept one connect & exit */
	    oneconnect = 1;			/* set one connect flag */
	    break;
	case 'u':
	case 'U':				/* use udp */
	    usetcp = 0;			/* reset tcp flag */
	    break;
	case 'p':
	case 'P':				/* port number */
	    portnum = atoi( optarg );
	    break;
	case '?':
	case 'h':
	case 'H':				/* help */
	default:
	    goto usage;			/* print out usage */
	}
    }
    if ( argc == optind )  return;
	    
usage:
    fprintf(stderr, "%s\n", what );
    fprintf(stderr, "Usage:  s2_xferd  <option>\n" );
    fprintf(stderr, " option:  -u   use UDP instead of TCP\n" );
    fprintf(stderr, "          -1   accept one connection then exit\n" );
#ifndef SINGLESERVER
    fprintf(stderr, "          -d   fork daemon and disconnect stdin\n" );
#endif
    fprintf(stderr, "          -pX  use service port number X\n" );
    fprintf(stderr, "          -h   help message\n" );
    fprintf(stderr, " example: s2_xferd \n" );
    fprintf(stderr, "          s2_xferd -u -p2000\n" );
#ifndef SINGLESERVER
    fprintf(stderr, "          s2_xferd -dup2000\n" );
#endif
    exit( 1 );
}
